import MuseScore 3.0
import QtQuick 2.5
import QtQuick.Dialogs 1.2
import QtQuick.Controls 2.1
import QtQuick.Layouts 1.3
import "Chords.js" as Chords

MuseScore 
{
  menuPath: "Plugins.pluginName"
  description: "Description goes here"
  version: "1.0"
  pluginType: "dialog"
  id: window
  width:  600;
  height: 500;
  
  onRun: 
  {
    scanChords();
  }
  
  Rectangle
  { 
    color: "lightGray"    
    anchors.fill: parent
  }

  Button 
  {
    id : okButton
    text: "OK"
    anchors.verticalCenter : window.verticalCenter 
    anchors.horizontalCenter : window.horizontalCenter 
    anchors.horizontalCenterOffset : -okButton.width
    onClicked: 
    { 
      writeChords();
      Qt.quit();  
    }
  }

  Button 
  {
    id : cancelButton
    text: "Cancel"
    anchors.verticalCenter : window.verticalCenter 
    anchors.horizontalCenter : window.horizontalCenter 
    anchors.horizontalCenterOffset : cancelButton.width
    onClicked: 
    { 
      Qt.quit();  
    }
  }
   
  property var pitchesTable : {"C":0,"Db":1,"D":2,"Eb":3,"E":4,"F":5,"Gb":6,"G":7,"Ab":8,"A":9,"Bb":10,"B":11};
  property var pitches : ["C","Db","D","Eb","E","F","Gb","G","Ab","A","Bb","B"];
  property var chordsStack : [];
  property var timesig;
   
  function writeChords()
  {
    var cursor = curScore.newCursor();
    cursor.track = curScore.ntracks - 4;
    clearStaff(cursor);
    cursor.rewind(Cursor.SCORE_START);
    timesig = cursor.measure.timesigNominal;
    writePickup(chordsStack[0], cursor);
    for (var i = 0; i < chordsStack.length ;i++)
      writeChord(chordsStack[i],cursor);
  }
   
  function writeChord(chord, cursor)
  {
    curScore.startCmd();
    cursor.setDuration(1,2);
    var pitch = chord.root.pitch;
    while (cursor.tick < chord.endTick)
        cursor.addNote(pitch);
    curScore.endCmd();    
  }
   
  function writePickup(chord, cursor)
  {
    var startTick = chordsStack[0].startTick;
    if (startTick == 0)
      return;
    curScore.startCmd();
    var timesigActual = cursor.measure.timesigActual;
    var beatDuration = 1920 / timesigActual.denominator;
    var pickupDuration = startTick / beatDuration;
    cursor.setDuration(1,pickupDuration);
    cursor.addRest();
    curScore.endCmd();    
  }
 
  // Remove all the notes in the given track.
  function clearStaff(cursor) 
  {
    curScore.startCmd();
    cursor.rewind(Cursor.SCORE_START);
    while (cursor.segment) 
    {
      if (cursor.element.type === Element.CHORD) 
        removeElement(cursor.element);
      cursor.next();
    }
    curScore.endCmd();    
}
  /*============================================================*/
   
  // Scan all the Harmony objects on the first staff,
  // create chords and fill the chords stack.
  function scanChords() 
  {
    var segment, tich;
    var cursor = curScore.newCursor();
    cursor.track = 0;
    cursor.rewind(cursor.SCORE_START);
    while (segment = cursor.segment) 
    {                   
      var text = getHarmony(segment);
      if (text)
      {
        var chord = createChord(cursor.tick,text);
        chordsStack.push(chord);
      }
      cursor.next();  
    }
    
    for (var i = 0; i < chordsStack.length - 1; i++)
      chordsStack[i].endTick = chordsStack[i + 1].startTick;
    chordsStack[chordsStack.length - 1].endTick = curScore.lastSegment.tick;
  }
 
  //Compute  diatonic interval between two notes.
  function subtractNote(note1,note2) 
  {
    var chr1 = note1.charCodeAt(0);
    var chr2 = note2.charCodeAt(0);
    var interval = chr2 - chr1;
    if (interval < 0)
      interval += 7;
    return interval
  }
  
  //Compute the chromatic interval between two notes.
  function subtractPitch(note1,note2) 
  {
    var pitch1 = computePitch(note1);
    var pitch2 = computePitch(note2);
    var interval = pitch2 - pitch1;
    return interval;
  }
  
  //Compute the basic pitch of a note.
  function computePitch(note) 
  {
    var pitch = pitchesTable[note[0]];
    if (note.length>1)
    {
      if(note[1]=='b')
        pitch--;
      if(note[1]=='#')
        pitch++;
    }
    if (note.length>2)
    {
      if(note[2]=='b')
        pitch--;
      if(note[2]=='#')
        pitch++;
    }
    return pitch;
  }
  
  // Create a note from a base note and a diatonic and Chromatic intervals.
  function getNote(baseNote,noteInterval,pitchInterval) 
  {
    var chr = baseNote.charCodeAt(0);
    chr += noteInterval;
    if (chr > "G".charCodeAt(0))
      chr -= 7;
    var basePitch = computePitch(baseNote);
    var computedPitch = basePitch + pitchInterval;
    var newNote = String.fromCharCode(chr);
    if(computedPitch > 11 && newNote != "B")
      computedPitch -= 12;
    var newPitch = pitchesTable[newNote];
    while (newPitch < computedPitch)
    {
      newNote += "#";
      newPitch++;
    }
    while (newPitch > computedPitch)
    {
      newNote += "b";
      newPitch--;
    }
    return newNote;
  }
  
  // Transpose C chord voicing to root voicing.
  function transposeChord(root,quality) 
  {
    var voicing = Chords.chordsTable[quality].split(' ');
    var chord = [root];
    for (var i = 1; i < voicing.length; i++)
    {
      var noteInterval = subtractNote(voicing[i-1],voicing[i]);
      var pitchInterval = subtractPitch(voicing[i-1],voicing[i]) 
      var newNote = getNote(chord[i-1],noteInterval,pitchInterval)
      chord.push(newNote);
    }
    return chord;
  }
  
  //Create a chord
  function createChord(tich,text) 
  {
    var step = text[0];
    var quality = "";
    if (text.length > 1)
    {
      if (text[1] == "b" || text[1] == "#")
        step += text[1];
      quality = text.replace(step,"");
    }
    
    var voicing = transposeChord(step,quality);
    var note = voicing[0];
    var pitch = computePitch(voicing[0])+ 36;
    var root = {note:note,pitch:pitch};
    var stack = [];
    for (var i in voicing)
    {
      note = voicing[i];
      pitch = computePitch(voicing[i])+ 48;
      var obj = {note:note,pitch:pitch};
      stack.push(obj);
    }
    stack.sort(function(a, b){return a.pitch - b.pitch});
    var chord = {startTick:tich,endTick:0,root:root,voicing:stack};
    return chord;
  }
  
  //Get all harmony objects on the first score
  function getHarmony(segment) 
  {
    if (segment.segmentType != Segment.ChordRest) 
      return null;
    var counter = 0;
    var annotation = segment.annotations[counter];
    var text, ch;
    while (annotation) 
    {
      if (annotation.type == Element.HARMONY)
      {
        text = annotation.text;
        ch = text[0];
        return text.replace(ch, ch.toUpperCase());
      }
      annotation = segment.annotations[++counter];   
    }
    return null;
  }
  
  function dispayProperties(object) 
  {
    if(!object)
    {
      console.log("null object");
      return;
    }
    console.log("---Dispay Properties---");
    var obj = Object.keys(object);
    for(var i=0 ;i<obj.length; i++)
    {
      var key = obj[i];
      console.log(key,"'", object[key], "'");
    }
    console.log("----------");
  }
}
   
  